/****************************************************************************
** Meta object code from reading C++ file 'ras_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../kimageformats-6.26.0/src/imageformats/ras_p.h"
#include <QtCore/qmetatype.h>
#include <QtCore/qplugin.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'ras_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN9RASPluginE_t {};
} // unnamed namespace

template <> constexpr inline auto RASPlugin::qt_create_metaobjectdata<qt_meta_tag_ZN9RASPluginE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "RASPlugin"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<RASPlugin, qt_meta_tag_ZN9RASPluginE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject RASPlugin::staticMetaObject = { {
    QMetaObject::SuperData::link<QImageIOPlugin::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN9RASPluginE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN9RASPluginE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN9RASPluginE_t>.metaTypes,
    nullptr
} };

void RASPlugin::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<RASPlugin *>(_o);
    (void)_t;
    (void)_c;
    (void)_id;
    (void)_a;
}

const QMetaObject *RASPlugin::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *RASPlugin::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN9RASPluginE_t>.strings))
        return static_cast<void*>(this);
    return QImageIOPlugin::qt_metacast(_clname);
}

int RASPlugin::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QImageIOPlugin::qt_metacall(_c, _id, _a);
    return _id;
}

#ifdef QT_MOC_EXPORT_PLUGIN_V2
static constexpr unsigned char qt_pluginMetaDataV2_RASPlugin[] = {
    0xbf, 
    // "IID"
    0x02,  0x78,  0x31,  'o',  'r',  'g',  '.',  'q', 
    't',  '-',  'p',  'r',  'o',  'j',  'e',  'c', 
    't',  '.',  'Q',  't',  '.',  'Q',  'I',  'm', 
    'a',  'g',  'e',  'I',  'O',  'H',  'a',  'n', 
    'd',  'l',  'e',  'r',  'F',  'a',  'c',  't', 
    'o',  'r',  'y',  'I',  'n',  't',  'e',  'r', 
    'f',  'a',  'c',  'e', 
    // "className"
    0x03,  0x69,  'R',  'A',  'S',  'P',  'l',  'u', 
    'g',  'i',  'n', 
    // "MetaData"
    0x04,  0xa2,  0x64,  'K',  'e',  'y',  's',  0x86, 
    0x63,  'i',  'm',  '1',  0x63,  'i',  'm',  '8', 
    0x64,  'i',  'm',  '2',  '4',  0x64,  'i',  'm', 
    '3',  '2',  0x63,  'r',  'a',  's',  0x63,  's', 
    'u',  'n',  0x69,  'M',  'i',  'm',  'e',  'T', 
    'y',  'p',  'e',  's',  0x86,  0x72,  'i',  'm', 
    'a',  'g',  'e',  '/',  'x',  '-',  's',  'u', 
    'n',  '-',  'r',  'a',  's',  't',  'e',  'r', 
    0x72,  'i',  'm',  'a',  'g',  'e',  '/',  'x', 
    '-',  's',  'u',  'n',  '-',  'r',  'a',  's', 
    't',  'e',  'r',  0x72,  'i',  'm',  'a',  'g', 
    'e',  '/',  'x',  '-',  's',  'u',  'n',  '-', 
    'r',  'a',  's',  't',  'e',  'r',  0x72,  'i', 
    'm',  'a',  'g',  'e',  '/',  'x',  '-',  's', 
    'u',  'n',  '-',  'r',  'a',  's',  't',  'e', 
    'r',  0x72,  'i',  'm',  'a',  'g',  'e',  '/', 
    'x',  '-',  's',  'u',  'n',  '-',  'r',  'a', 
    's',  't',  'e',  'r',  0x72,  'i',  'm',  'a', 
    'g',  'e',  '/',  'x',  '-',  's',  'u',  'n', 
    '-',  'r',  'a',  's',  't',  'e',  'r', 
    0xff, 
};
QT_MOC_EXPORT_PLUGIN_V2(RASPlugin, RASPlugin, qt_pluginMetaDataV2_RASPlugin)
#else
QT_PLUGIN_METADATA_SECTION
Q_CONSTINIT static constexpr unsigned char qt_pluginMetaData_RASPlugin[] = {
    'Q', 'T', 'M', 'E', 'T', 'A', 'D', 'A', 'T', 'A', ' ', '!',
    // metadata version, Qt version, architectural requirements
    0, QT_VERSION_MAJOR, QT_VERSION_MINOR, qPluginArchRequirements(),
    0xbf, 
    // "IID"
    0x02,  0x78,  0x31,  'o',  'r',  'g',  '.',  'q', 
    't',  '-',  'p',  'r',  'o',  'j',  'e',  'c', 
    't',  '.',  'Q',  't',  '.',  'Q',  'I',  'm', 
    'a',  'g',  'e',  'I',  'O',  'H',  'a',  'n', 
    'd',  'l',  'e',  'r',  'F',  'a',  'c',  't', 
    'o',  'r',  'y',  'I',  'n',  't',  'e',  'r', 
    'f',  'a',  'c',  'e', 
    // "className"
    0x03,  0x69,  'R',  'A',  'S',  'P',  'l',  'u', 
    'g',  'i',  'n', 
    // "MetaData"
    0x04,  0xa2,  0x64,  'K',  'e',  'y',  's',  0x86, 
    0x63,  'i',  'm',  '1',  0x63,  'i',  'm',  '8', 
    0x64,  'i',  'm',  '2',  '4',  0x64,  'i',  'm', 
    '3',  '2',  0x63,  'r',  'a',  's',  0x63,  's', 
    'u',  'n',  0x69,  'M',  'i',  'm',  'e',  'T', 
    'y',  'p',  'e',  's',  0x86,  0x72,  'i',  'm', 
    'a',  'g',  'e',  '/',  'x',  '-',  's',  'u', 
    'n',  '-',  'r',  'a',  's',  't',  'e',  'r', 
    0x72,  'i',  'm',  'a',  'g',  'e',  '/',  'x', 
    '-',  's',  'u',  'n',  '-',  'r',  'a',  's', 
    't',  'e',  'r',  0x72,  'i',  'm',  'a',  'g', 
    'e',  '/',  'x',  '-',  's',  'u',  'n',  '-', 
    'r',  'a',  's',  't',  'e',  'r',  0x72,  'i', 
    'm',  'a',  'g',  'e',  '/',  'x',  '-',  's', 
    'u',  'n',  '-',  'r',  'a',  's',  't',  'e', 
    'r',  0x72,  'i',  'm',  'a',  'g',  'e',  '/', 
    'x',  '-',  's',  'u',  'n',  '-',  'r',  'a', 
    's',  't',  'e',  'r',  0x72,  'i',  'm',  'a', 
    'g',  'e',  '/',  'x',  '-',  's',  'u',  'n', 
    '-',  'r',  'a',  's',  't',  'e',  'r', 
    0xff, 
};
QT_MOC_EXPORT_PLUGIN(RASPlugin, RASPlugin)
#endif  // QT_MOC_EXPORT_PLUGIN_V2

QT_WARNING_POP
